@@ -4,15 +4,9 @@
 #ifndef DIFF_H
 #define DIFF_H
 
-#define DIFF_FILE_CANON_MODE(mode) \
-	(S_ISREG(mode) ? (S_IFREG | ce_permissions(mode)) : \
-	S_ISLNK(mode) ? S_IFLNK : S_IFDIR)
-
-struct tree_desc {
-	void *buf;
-	unsigned long size;
-};
+#include "tree-walk.h"
 
+struct rev_info;
 struct diff_options;
 
 typedef void (*change_fn_t)(struct diff_options *options,
@@ -27,16 +21,22 @@ typedef void (*add_remove_fn_t)(struct diff_options *options,
 		    const char *base, const char *path);
 
 struct diff_options {
-	const char **paths;
 	const char *filter;
 	const char *orderfile;
 	const char *pickaxe;
 	unsigned recursive:1,
+		 with_raw:1,
+		 with_stat:1,
 		 tree_in_recursive:1,
-		 full_index:1;
+		 binary:1,
+		 full_index:1,
+		 silent_on_remove:1,
+		 find_copies_harder:1,
+		 summary:1,
+		 color_diff:1;
+	int context;
 	int break_opt;
 	int detect_rename;
-	int find_copies_harder;
 	int line_termination;
 	int output_format;
 	int pickaxe_opts;
@@ -45,17 +45,48 @@ struct diff_options {
 	int rename_limit;
 	int setup;
 	int abbrev;
+	const char *stat_sep;
+	long xdl_opts;
 
+	int nr_paths;
+	const char **paths;
+	int *pathlens;
 	change_fn_t change;
 	add_remove_fn_t add_remove;
 };
 
-extern void diff_tree_setup_paths(const char **paths);
+extern const char mime_boundary_leader[];
+
+extern void diff_tree_setup_paths(const char **paths, struct diff_options *);
+extern void diff_tree_release_paths(struct diff_options *);
 extern int diff_tree(struct tree_desc *t1, struct tree_desc *t2,
 		     const char *base, struct diff_options *opt);
 extern int diff_tree_sha1(const unsigned char *old, const unsigned char *new,
 			  const char *base, struct diff_options *opt);
 
+struct combine_diff_path {
+	struct combine_diff_path *next;
+	int len;
+	char *path;
+	unsigned int mode;
+	unsigned char sha1[20];
+	struct combine_diff_parent {
+		char status;
+		unsigned int mode;
+		unsigned char sha1[20];
+	} parent[FLEX_ARRAY];
+};
+#define combine_diff_path_size(n, l) \
+	(sizeof(struct combine_diff_path) + \
+	 sizeof(struct combine_diff_parent) * (n) + (l) + 1)
+
+extern void show_combined_diff(struct combine_diff_path *elem, int num_parent,
+			      int dense, struct rev_info *);
+
+extern void diff_tree_combined(const unsigned char *sha1, const unsigned char parent[][20], int num_parent, int dense, struct rev_info *rev);
+
+extern void diff_tree_combined_merge(const unsigned char *sha1, int, struct rev_info *);
+
 extern void diff_addremove(struct diff_options *,
 			   int addremove,
 			   unsigned mode,
@@ -87,9 +118,7 @@ extern int diff_setup_done(struct diff_options *);
 #define DIFF_DETECT_COPY	2
 
 #define DIFF_PICKAXE_ALL	1
-
-#define DIFF_DEFAULT_INDEX_ABBREV	7 /* hex digits */
-#define DIFF_DEFAULT_ABBREV	7 /* hex digits */
+#define DIFF_PICKAXE_REGEX	2
 
 extern void diffcore_std(struct diff_options *);
 
@@ -100,6 +129,11 @@ extern void diffcore_std_no_resolve(struct diff_options *);
 "  -z            output diff-raw with lines terminated with NUL.\n" \
 "  -p            output patch format.\n" \
 "  -u            synonym for -p.\n" \
+"  --patch-with-raw\n" \
+"                output both a patch and the diff-raw format.\n" \
+"  --stat        show diffstat instead of patch.\n" \
+"  --patch-with-stat\n" \
+"                output a patch and prepend its diffstat.\n" \
 "  --name-only   show only names of changed files.\n" \
 "  --name-status show names and status of changed files.\n" \
 "  --full-index  show full object name on index lines.\n" \
@@ -123,6 +157,8 @@ extern int diff_queue_is_empty(void);
 #define DIFF_FORMAT_NO_OUTPUT	3
 #define DIFF_FORMAT_NAME	4
 #define DIFF_FORMAT_NAME_STATUS	5
+#define DIFF_FORMAT_DIFFSTAT	6
+#define DIFF_FORMAT_CHECKDIFF	7
 
 extern void diff_flush(struct diff_options*);
 
@@ -144,4 +180,10 @@ extern void diff_flush(struct diff_options*);
 
 extern const char *diff_unique_abbrev(const unsigned char *, int);
 
+extern int run_diff_files(struct rev_info *revs, int silent_on_removed);
+
+extern int run_diff_index(struct rev_info *revs, int cached);
+
+extern int diff_flush_patch_id(struct diff_options *, unsigned char *);
+
 #endif /* DIFF_H */
