@@ -4,27 +4,53 @@
 struct object_list {
 	struct object *item;
 	struct object_list *next;
-	const char *name;
 };
 
 struct object_refs {
 	unsigned count;
+	struct object *base;
 	struct object *ref[FLEX_ARRAY]; /* more */
 };
 
+struct object_array {
+	unsigned int nr;
+	unsigned int alloc;
+	struct object_array_entry {
+		struct object *item;
+		const char *name;
+	} *objects;
+};
+
+#define TYPE_BITS   3
+#define FLAG_BITS  27
+
+#define TYPE_NONE   0
+#define TYPE_BLOB   1
+#define TYPE_TREE   2
+#define TYPE_COMMIT 3
+#define TYPE_TAG    4
+#define TYPE_BAD    5
+
 struct object {
 	unsigned parsed : 1;
 	unsigned used : 1;
-	unsigned int flags;
+	unsigned type : TYPE_BITS;
+	unsigned flags : FLAG_BITS;
 	unsigned char sha1[20];
-	const char *type;
-	struct object_refs *refs;
-	void *util;
 };
 
 extern int track_object_refs;
-extern int nr_objs;
-extern struct object **objs;
+extern const char *type_names[];
+
+extern unsigned int get_max_object_index(void);
+extern struct object *get_indexed_object(unsigned int);
+
+static inline const char *typename(unsigned int type)
+{
+	return type_names[type > TYPE_TAG ? TYPE_BAD : type];
+}
+
+extern struct object_refs *lookup_object_refs(struct object *);
 
 /** Internal only **/
 struct object *lookup_object(const unsigned char *sha1);
@@ -55,4 +81,7 @@ unsigned object_list_length(struct object_list *list);
 
 int object_list_contains(struct object_list *list, struct object *obj);
 
+/* Object array handling .. */
+void add_object_array(struct object *obj, const char *name, struct object_array *array);
+
 #endif /* OBJECT_H */
