@@ -3,6 +3,8 @@
 #include "commit.h"
 #include "tree.h"
 #include "blob.h"
+#include "tree-walk.h"
+#include "refs.h"
 
 static int find_short_object_filename(int len, const char *name, unsigned char *sha1)
 {
@@ -143,7 +145,7 @@ static int find_unique_short_object(int len, char *canonical,
 	}
 	/* Both have unique ones -- do they match? */
 	if (memcmp(packed_sha1, unpacked_sha1, 20))
-		return -2;
+		return SHORT_NAME_AMBIGUOUS;
 	memcpy(sha1, packed_sha1, 20);
 	return 0;
 }
@@ -155,7 +157,7 @@ static int get_short_sha1(const char *name, int len, unsigned char *sha1,
 	char canonical[40];
 	unsigned char res[20];
 
-	if (len < 4)
+	if (len < MINIMUM_ABBREV)
 		return -1;
 	memset(res, 0, 20);
 	memset(canonical, 'x', 40);
@@ -186,16 +188,18 @@ static int get_short_sha1(const char *name, int len, unsigned char *sha1,
 
 const char *find_unique_abbrev(const unsigned char *sha1, int len)
 {
-	int status;
+	int status, is_null;
 	static char hex[41];
 
+	is_null = !memcmp(sha1, null_sha1, 20);
 	memcpy(hex, sha1_to_hex(sha1), 40);
 	if (len == 40)
 		return hex;
 	while (len < 40) {
 		unsigned char sha1_ret[20];
 		status = get_short_sha1(hex, len, sha1_ret, 1);
-		if (!status) {
+		if (!status ||
+		    (is_null && status != SHORT_NAME_AMBIGUOUS)) {
 			hex[len] = 0;
 			return hex;
 		}
@@ -233,28 +237,69 @@ static int ambiguous_path(const char *path, int len)
 
 static int get_sha1_basic(const char *str, int len, unsigned char *sha1)
 {
-	static const char *prefix[] = {
-		"",
-		"refs",
-		"refs/tags",
-		"refs/heads",
+	static const char *fmt[] = {
+		"%.*s",
+		"refs/%.*s",
+		"refs/tags/%.*s",
+		"refs/heads/%.*s",
+		"refs/remotes/%.*s",
+		"refs/remotes/%.*s/HEAD",
 		NULL
 	};
-	const char **p;
+	static const char *warning = "warning: refname '%.*s' is ambiguous.\n";
+	const char **p, *pathname;
+	char *real_path = NULL;
+	int refs_found = 0, am;
+	unsigned long at_time = (unsigned long)-1;
+	unsigned char *this_result;
+	unsigned char sha1_from_ref[20];
 
 	if (len == 40 && !get_sha1_hex(str, sha1))
 		return 0;
 
+	/* At a given period of time? "@{2 hours ago}" */
+	for (am = 1; am < len - 1; am++) {
+		if (str[am] == '@' && str[am+1] == '{' && str[len-1] == '}') {
+			int date_len = len - am - 3;
+			char *date_spec = xmalloc(date_len + 1);
+			strlcpy(date_spec, str + am + 2, date_len + 1);
+			at_time = approxidate(date_spec);
+			free(date_spec);
+			len = am;
+			break;
+		}
+	}
+
 	/* Accept only unambiguous ref paths. */
 	if (ambiguous_path(str, len))
 		return -1;
 
-	for (p = prefix; *p; p++) {
-		char *pathname = git_path("%s/%.*s", *p, len, str);
-		if (!read_ref(pathname, sha1))
-			return 0;
+	for (p = fmt; *p; p++) {
+		this_result = refs_found ? sha1_from_ref : sha1;
+		pathname = resolve_ref(git_path(*p, len, str), this_result, 1);
+		if (pathname) {
+			if (!refs_found++)
+				real_path = strdup(pathname);
+			if (!warn_ambiguous_refs)
+				break;
+		}
 	}
-	return -1;
+
+	if (!refs_found)
+		return -1;
+
+	if (warn_ambiguous_refs && refs_found > 1)
+		fprintf(stderr, warning, len, str);
+
+	if (at_time != (unsigned long)-1) {
+		read_ref_at(
+			real_path + strlen(git_path(".")) - 1,
+			at_time,
+			sha1);
+	}
+
+	free(real_path);
+	return 0;
 }
 
 static int get_sha1_1(const char *name, int len, unsigned char *sha1);
@@ -312,7 +357,7 @@ static int peel_onion(const char *name, int len, unsigned char *sha1)
 {
 	unsigned char outer[20];
 	const char *sp;
-	const char *type_string = NULL;
+	unsigned int expected_type = 0;
 	struct object *o;
 
 	/*
@@ -336,13 +381,13 @@ static int peel_onion(const char *name, int len, unsigned char *sha1)
 
 	sp++; /* beginning of type name, or closing brace for empty */
 	if (!strncmp(commit_type, sp, 6) && sp[6] == '}')
-		type_string = commit_type;
+		expected_type = TYPE_COMMIT;
 	else if (!strncmp(tree_type, sp, 4) && sp[4] == '}')
-		type_string = tree_type;
+		expected_type = TYPE_TREE;
 	else if (!strncmp(blob_type, sp, 4) && sp[4] == '}')
-		type_string = blob_type;
+		expected_type = TYPE_BLOB;
 	else if (sp[0] == '}')
-		type_string = NULL;
+		expected_type = TYPE_NONE;
 	else
 		return -1;
 
@@ -352,7 +397,7 @@ static int peel_onion(const char *name, int len, unsigned char *sha1)
 	o = parse_object(outer);
 	if (!o)
 		return -1;
-	if (!type_string) {
+	if (!expected_type) {
 		o = deref_tag(o, name, sp - name - 2);
 		if (!o || (!o->parsed && !parse_object(o->sha1)))
 			return -1;
@@ -367,18 +412,18 @@ static int peel_onion(const char *name, int len, unsigned char *sha1)
 		while (1) {
 			if (!o || (!o->parsed && !parse_object(o->sha1)))
 				return -1;
-			if (o->type == type_string) {
+			if (o->type == expected_type) {
 				memcpy(sha1, o->sha1, 20);
 				return 0;
 			}
-			if (o->type == tag_type)
+			if (o->type == TYPE_TAG)
 				o = ((struct tag*) o)->tagged;
-			else if (o->type == commit_type)
+			else if (o->type == TYPE_COMMIT)
 				o = &(((struct commit *) o)->tree->object);
 			else
 				return error("%.*s: expected %s type, but the object dereferences to %s type",
-					     len, name, type_string,
-					     o->type);
+					     len, name, typename(expected_type),
+					     typename(o->type));
 			if (!o->parsed)
 				parse_object(o->sha1);
 		}
@@ -388,43 +433,36 @@ static int peel_onion(const char *name, int len, unsigned char *sha1)
 
 static int get_sha1_1(const char *name, int len, unsigned char *sha1)
 {
-	int parent, ret;
+	int ret, has_suffix;
 	const char *cp;
 
-	/* foo^[0-9] or foo^ (== foo^1); we do not do more than 9 parents. */
-	if (len > 2 && name[len-2] == '^' &&
-	    name[len-1] >= '0' && name[len-1] <= '9') {
-		parent = name[len-1] - '0';
-		len -= 2;
-	}
-	else if (len > 1 && name[len-1] == '^') {
-		parent = 1;
-		len--;
-	} else
-		parent = -1;
-
-	if (parent >= 0)
-		return get_parent(name, len, sha1, parent);
-
 	/* "name~3" is "name^^^",
-	 * "name~12" is "name^^^^^^^^^^^^", and
 	 * "name~" and "name~0" are name -- not "name^0"!
+	 * "name^" is not "name^0"; it is "name^1".
 	 */
-	parent = 0;
+	has_suffix = 0;
 	for (cp = name + len - 1; name <= cp; cp--) {
 		int ch = *cp;
 		if ('0' <= ch && ch <= '9')
 			continue;
-		if (ch != '~')
-			parent = -1;
+		if (ch == '~' || ch == '^')
+			has_suffix = ch;
 		break;
 	}
-	if (!parent && *cp == '~') {
+
+	if (has_suffix) {
+		int num = 0;
 		int len1 = cp - name;
 		cp++;
 		while (cp < name + len)
-			parent = parent * 10 + *cp++ - '0';
-		return get_nth_ancestor(name, len1, sha1, parent);
+			num = num * 10 + *cp++ - '0';
+		if (has_suffix == '^') {
+			if (!num && len1 == len - 1)
+				num = 1;
+			return get_parent(name, len1, sha1, num);
+		}
+		/* else if (has_suffix == '~') -- goes without saying */
+		return get_nth_ancestor(name, len1, sha1, num);
 	}
 
 	ret = peel_onion(name, len, sha1);
@@ -443,6 +481,65 @@ static int get_sha1_1(const char *name, int len, unsigned char *sha1)
  */
 int get_sha1(const char *name, unsigned char *sha1)
 {
+	int ret, bracket_depth;
+	unsigned unused;
+	int namelen = strlen(name);
+	const char *cp;
+
 	prepare_alt_odb();
-	return get_sha1_1(name, strlen(name), sha1);
+	ret = get_sha1_1(name, namelen, sha1);
+	if (!ret)
+		return ret;
+	/* sha1:path --> object name of path in ent sha1
+	 * :path -> object name of path in index
+	 * :[0-3]:path -> object name of path in index at stage
+	 */
+	if (name[0] == ':') {
+		int stage = 0;
+		struct cache_entry *ce;
+		int pos;
+		if (namelen < 3 ||
+		    name[2] != ':' ||
+		    name[1] < '0' || '3' < name[1])
+			cp = name + 1;
+		else {
+			stage = name[1] - '0';
+			cp = name + 3;
+		}
+		namelen = namelen - (cp - name);
+		if (!active_cache)
+			read_cache();
+		if (active_nr < 0)
+			return -1;
+		pos = cache_name_pos(cp, namelen);
+		if (pos < 0)
+			pos = -pos - 1;
+		while (pos < active_nr) {
+			ce = active_cache[pos];
+			if (ce_namelen(ce) != namelen ||
+			    memcmp(ce->name, cp, namelen))
+				break;
+			if (ce_stage(ce) == stage) {
+				memcpy(sha1, ce->sha1, 20);
+				return 0;
+			}
+			pos++;
+		}
+		return -1;
+	}
+	for (cp = name, bracket_depth = 0; *cp; cp++) {
+		if (*cp == '{')
+			bracket_depth++;
+		else if (bracket_depth && *cp == '}')
+			bracket_depth--;
+		else if (!bracket_depth && *cp == ':')
+			break;
+	}
+	if (*cp == ':') {
+		unsigned char tree_sha1[20];
+		if (!get_sha1_1(name, cp-name, tree_sha1))
+			return get_tree_entry(tree_sha1, cp+1, sha1,
+					      &unused);
+	}
+	return ret;
 }
